import cgi

from aspen.http.mapping import Mapping


class WwwForm(Mapping):
    """Represent a WWW form.
    """

    def __init__(self, s):
        """Takes a string of type application/x-www-form-urlencoded.
        """
        self.raw = s
        self._dict = cgi.parse_qs( s
                                 , keep_blank_values = True
                                 , strict_parsing = False
                                  )



class Mapping(object):
    """Base class for HTTP mappings.

    HTTP forms and headers may have a single item or a list of items as the
    value. So while Python dictionary semantics work for almost everything, it
    is better (IMO) for the API to force you to be explicit about whether you
    are expecting a single item or list of items. We do that here by providing
    'one' and 'all' methods, rather than item access and a 'get' method.
    Furthermore, this class supports iteration over keys, but not iteration
    over values. Iterate over keys, and then use one or all.

    All API here operates on a self._dict dictionary. Set that in subclass
    constructors.

    """

    def __init__(self, **kw):
        self._dict = {}
        for name, value in kw.iteritems():
            self.set(name, value)

    def add(self, name, value):
        """Given a name and value, add another entry.
        """
        if name not in self._dict:
            self.set(name, value)
        else:
            self._dict[name].append(value)

    def __contains__(self, name):
        return name in self._dict

    def all(self, name, default=None):
        """Given a name, return a list of values.
        """
        if default is None:
            default = []
        return self._dict.get(name, default)
       
    def one(self, name, default=None):
        return self._dict.get(name, [default])[0]

    def ones(self, *names):
        """Given one or more names of keys, return a list of their values.
        """
        return [self.one(name) for name in names]

    def __iter__(self):
        return self._dict

    def __in__(self, name):
        """Given a name, return True if it is known in the mapping.
        """
        return name in self._dict

    def __iter__(self):
        return self._dict.__iter__()

    def keys(self):
        """Return a list of names.
        """
        return self._dict.keys()

    def set(self, name, value):
        """Given a name and value, set the value, clearing all others.
        
        Pass None to remove.

        """
        if value is None:
            del self._dict[name]
        self._dict[name] = [str(value).strip()] # TODO unicode?


    # Convenience methods for coercing to bool.
    # =========================================

    def yes(self, name):
        """Given a key name, return a boolean.
        
        The value for the key must be in the set {0,1,yes,no,true,false},
        case-insensistive. If the key is not in this section, we return True.

        """
        return self._yes_no(name, True)

    def no(self, name):
        """Given a key name, return a boolean.
        
        The value for the key must be in the set {0,1,yes,no,true,false},
        case-insensistive. If the key is not in this section, we return False.

        """
        return self._yes_no(name, False)

    def _yes_no(self, name, default):
        if name not in self._dict:
            return default 
        value = self._dict[name].lower()
        if value not in YES_NO:
            raise ConfigurationError( "%s should be 'yes' or 'no', not %s" 
                                    % (name, self._dict[name])
                                     )
        return value in YES 





import datetime


# Find a json module.
# ===================
# The standard library includes simplejson as json since 2.6, but without the 
# C speedups. So we prefer simplejson if it is available.

try:
    import simplejson as _json
except ImportError:
    try:
        import json as _json
    except ImportError:
        _json = None


# Support pymongo binary #redhotbaby.
# ===================================

try:
    from pymongo.binary import Binary
except ImportError:
    class Binary:
        pass

if _json is not None:
    class FriendlyEncoder(_json.JSONEncoder):
        """Add support for additional types to the default JSON encoder.
        """

        def default(self, obj):
            if isinstance(obj, complex):
                # http://docs.python.org/library/json.html
                out = [obj.real, obj.imag]
            elif isinstance(obj, datetime.datetime):
                # http://stackoverflow.com/questions/455580/
                out = obj.isoformat()
            elif isinstance(obj, Binary):
                out = str(obj).encode('base64')
            else:
                out = super(FriendlyEncoder, self).default(obj)
            return out

def lazy_check(): 
    if _json is None:
        raise LoadError("Neither json nor simplejson was found. Try "
                        "installing simplejson to use dynamic JSON "
                        "resources. See "
                        "http://aspen.io/resources/json/#libraries for "
                        "more information.")

def loads(*a, **kw):
    lazy_check()
    return _json.loads(*a, **kw)

def dumps(*a, **kw):
    lazy_check()
    if 'cls' not in kw:
        kw['cls'] = FriendlyEncoder
    return _json.dumps(*a, **kw)




import diesel
from aspen.engines import CooperativeEngine
from diesel.protocols import wsgi


class Engine(CooperativeEngine):

    diesel_app = None # a diesel diesel_app instance

    def bind(self):
        self.diesel_app = wsgi.WSGIApplication( self.website
                                              , self.website.address[1]
                                              , self.website.address[0]
                                               )

    def start(self):
        self.diesel_app.run()

    def stop(self):
        try:
            if self.diesel_app is not None:
                self.diesel_app.halt()
        except diesel.app.ApplicationEnd:
            pass # Only you can prevent log spam.

    def start_restarter(self, check_all):
        def loop():
            while True:
                check_all()
                diesel.sleep(0.5)
        self.diesel_app.add_loop(diesel.Loop(loop))



def _authorize_anonymous(path):
    """Given the path part of an URL, return a boolean.
    """
    if path in ('/favicon.ico', '/robots.txt'): # special cases
        return True
    if path and path.startswith('/anonymous/'): # logging in
        return True
    return False

def inbound(request):
    """Authenticate from a cookie.
    """
    session = {}
    if 'session_id' in request.cookie:
        session_id = request.cookie['session_id'].value
        session = sessions.get(session_id, {})

    request.user = User(session)
    if not session:
        if not _authorize_anonymous(request.path.raw):
            raise Response(401) # use nice error messages for login form

def outbound(response):
    session = response.request.user.session
    if not session:                                 # user is anonymous
        if 'session_id' not in response.request.cookie:
            # no cookie in the request, don't set one on response
            return
        else:
            # expired cookie in the request, instruct browser to delete it
            response.cookie['session_id'] = '' 
            expires = 0
    else:                                           # user is authenticated
        response.headers.set('Expires', BEGINNING_OF_EPOCH) # don't cache
        response.cookie['session_id'] = session['_id']
        expires = session['expires'] = time.time() + TIMEOUT

    cookie = response.cookie['session_id']
    # I am not setting domain, because it is supposed to default to what we 
    # want: the domain of the object requested.
    #cookie['domain']
    cookie['path'] = '/'
    cookie['expires'] = rfc822.formatdate(expires)
    cookie['httponly'] = "Yes, please."

def startup(website):
    """Read in configuration.
    """
    global backend
    log.info('aspen.authentication.startup')
 
    
    # Backend
    # =======
    # For development offline we provide a stub backend that doesn't hit LDAP.

    if website.conf.aspen['auth_backend'] is MissedConnection:
        log.warning("auth backend not configured, using stub authentication")
        backend = StubBackend()
    else:
        log.info("ldap configured, using normal authentication")
        backend = ldap_

    
    # Start sync thread.
    # ==================
    # Once primed, we have an async thread to keep mongo in sync. So if we go
    # down hard, we lose a little bit of session state, plus whatever mongo
    # hadn't flushed to disk in time. Since we assume a one to one between http
    # and mongo, we only *read* from mongo on startup.

    spec = {'_id': {'$nin': sessions.keys()}}
    for session in dbs.ours.sessions.find(spec):
        sessions[session['_id']] = session

    Periodic(flush_sessions, seconds=5).start()
